#include <Queue.h>
#include <FastLED.h>
#include <Streaming.h>
#include "Menu.h"
#include "Encoder.h"
#include "Effects.h"

/* hardware configurations */
#define TEENSY (1)

#define COLOR_ORDER GRB
#define CHIPSET     WS2812
#define NUM_LEDS    29
#define COLOR_CORRECTION 0xe0ffe0
#define BRIGHTNESS  140

#if TEENSY
#define ENC_PULSE_PIN             6
#define ENC_STEP_PIN              5
#define ENC_DIR_PIN               4
#define LC_DATA_PIN               3
#define LC_CLK_PIN                2
#define LC_CS_PIN                 1
#define LED_BALL_PIN              0
#define LEDPIN                   11
#else
#define ENC_PULSE_PIN             2
#define ENC_STEP_PIN              3
#define ENC_DIR_PIN               4
#define LC_DATA_PIN               7
#define LC_CLK_PIN                5
#define LC_CS_PIN                 6
#define LED_BALL_PIN              9
#define LEDPIN                   13
#endif


/* DEBUG CONFIGURATION */
#define DEBUG (1)
#define DEBUG_BLINK   (1 & DEBUG)
#define DEBUG_EFFECTS (0 & DEBUG)
#define DEBUG_ENCODER (1 & DEBUG)
#define DEBUG_MENU    (1 & DEBUG)

/* using of palettes */
#define USE_PALETTES 1

CRGB leds[NUM_LEDS];

typedef struct {uint8_t pixelOffset;
                uint8_t pixelsInBand;
} LEDSlice_t;

typedef struct {
  const char* id;
  queuedFunction function;
  uint8_t delay;
  void* userData;
}effectProgram_t;

CRGBPalette16 colorPalettes[]={
  RainbowColors_p,
  CloudColors_p,
  LavaColors_p,
  OceanColors_p,
  ForestColors_p,
  RainbowStripeColors_p,
  PartyColors_p,
  HeatColors_p
};

uint8_t numberOfPalettes = sizeof(colorPalettes)/sizeof(CRGBPalette16);

int16_t bandSegments[]={1,255/3,30,50,100};
effectProgram_t effectPrograms[]  ={
  {"FANB",fadeNeighbours,50,NULL},
  {"WRDD",runningDots,32,NULL},
  {"band",rainbowBands,80,NULL},                          // default, 5 colors in ball
//  {"band",rainbowBands,80,&bandSegments[1]},
//  {"band",rainbowBands,80,&bandSegments[2]},
//  {"band",rainbowBands,80,&bandSegments[3]},
//  {"band",rainbowBands,80,&bandSegments[4]},
  {"spir",rainbowSpiral,80,NULL},
  {"firr",Fire2012,30,NULL},
};

uint8_t maxPrograms = sizeof(effectPrograms) / sizeof(effectProgram_t);

LEDSlice_t bands[5] = {
  {0,6},
  {6,8},
  {14,8},
  {22,6},
  {28,1}
};

Queue taskQueue;
ballState_t currentState;
int16_t currentDelay = 10;
volatile int16_t  nextDelay = currentDelay;
int16_t currentProgram;
volatile int16_t nextProgram;
int16_t currentPalette = 0;
volatile int16_t nextPalette;
int16_t currentBrightness;
volatile int16_t nextBrightness;

boolean effectStarted = false;

#if DEBUG_BLINK
int blinker(unsigned long now, void* userData)
{
  static boolean state = HIGH;
  digitalWrite(LEDPIN,state);
  state = !state;
}
#endif

void setup()
{
  delay(200);
  #if DEBUG
    Serial.begin(57600);
    Serial <<" Setup "<<endl;
    Serial <<" Programs " << maxPrograms<<endl;
    Serial <<" Palettes " << numberOfPalettes<<endl;
    Serial <<" LEDPIN "<<LEDPIN<<endl;
    Serial <<" Neightbours:"<<sizeof(neightbours_t)<<" "<<neightbours[0].topLeftNeightbour<<endl;
  #endif

  #if DEBUG_BLINK
    pinMode(LEDPIN, OUTPUT);
  #endif

  FastLED.addLeds<CHIPSET, LED_BALL_PIN, COLOR_ORDER>(leds,NUM_LEDS).setCorrection( COLOR_CORRECTION );
//  FastLED.setColor
  FastLED.setBrightness(BRIGHTNESS);
  FastLED.setDither(0);
  FastLED.clear(true);
  #if DEBUG
    Serial << "fastLed Ready"<<endl;
  #endif
  currentState = stateNone;
  initEncoderUI();

  #if DEBUG
    Serial << "Encoder ready"<<endl;
  #endif

  #if DEBUG_BLINK
    taskQueue.scheduleFunction(blinker,NULL,"BLNK",1000,1000);
    Serial << "Blinker Running"<< endl;
  #endif

  currentProgram = nextProgram = 0;
  currentDelay = nextDelay = effectPrograms[currentProgram].delay;
  currentBrightness = nextBrightness = BRIGHTNESS;
  currentPalette = nextPalette = 0;
  #if DEBUG
    Serial << "Program init"<<endl;
  #endif
  taskQueue.scheduleFunction(effectPrograms[currentProgram].function,effectPrograms[currentProgram].userData,effectPrograms[currentProgram].id,0,currentDelay);
  effectStarted = true;
  encoderClickCallback = clickHandler;

  lc.clearDisplay(0);
  randomSeed(millis());
  #if DEBUG
    Serial << "Setup finished"<<endl;
  #endif
}

void loop()
{
	//Serial << "l."<<endl;
  if(nextProgram != currentProgram){          // switch program
#if DEBUG
    Serial << currentProgram<<" "<<nextProgram<<" ";
#endif
    effectStarted = true;
    taskQueue.scheduleRemoveFunction(effectPrograms[currentProgram].id);
    taskQueue.scheduleFunction(effectPrograms[nextProgram].function,effectPrograms[nextProgram].userData,effectPrograms[nextProgram].id,currentDelay,currentDelay);
    nextDelay=currentDelay =effectPrograms[nextProgram].delay;
    currentProgram = nextProgram;
    #if DEBUG
    Serial << "PQueue" << taskQueue._itemsInQueue<<endl;
    #endif
  }
  // switch delay
  if(currentDelay != nextDelay){
    #if DEBUG
    Serial<<currentDelay<<" "<<nextDelay<<" ";
    #endif
      taskQueue.scheduleChangeFunction(effectPrograms[currentProgram].id,nextDelay,nextDelay);
      currentDelay = nextDelay;
      #if DEBUG
      Serial << "DQueue" << taskQueue._itemsInQueue<<endl;
      #endif
  }
  // switch brightness
  if(currentBrightness != nextBrightness){
    #if DEBUG
    Serial <<"B"<<endl;
    #endif
    FastLED.setBrightness(nextBrightness);
    currentBrightness = nextBrightness;
  }
  // switch palette
  if(currentPalette != nextPalette){
    currentPalette = nextPalette;
  }
  taskQueue.Run(millis());
  delay(1);
//  Serial << "Loop end"<<endl;
}
